import inspect
import logging
import multiprocessing
import os
import shutil
import sys
import tempfile
import threading
import time
from io import TextIOWrapper
from logging import Logger
from queue import Queue
from types import MethodType
from typing import Any, Optional, List, Union, ContextManager

from larakit import StatefulNamespace


def mp_apply(generator, fn, *, pool_init=None, pool_init_args=None,
             batch_size: int = 1, ordered: bool = True, threads: int = None):
    cores = threads or multiprocessing.cpu_count()

    def loader(generator_, jobs_):
        for args in generator_:
            jobs_.put(args, block=True)
        jobs_.put(None, block=True)

    jobs = Queue(maxsize=cores * batch_size)
    loader_thread = threading.Thread(target=loader, args=(generator, jobs), daemon=True)

    loader_thread.start()

    with multiprocessing.Pool(initializer=pool_init, initargs=pool_init_args, processes=cores) as pool:
        imap_f = pool.imap if ordered else pool.imap_unordered

        yield from imap_f(fn, iter(jobs.get, None), chunksize=batch_size)

    loader_thread.join()


def mp_stream(generator, fn, batch_size=1, threads=None, queue_size=None):
    threads = threads or multiprocessing.cpu_count()
    queue_size = queue_size or threads * batch_size

    in_q = multiprocessing.Queue(maxsize=queue_size)
    out_q = multiprocessing.Queue(maxsize=queue_size * 2)

    def worker(iq: multiprocessing.Queue, oq: multiprocessing.Queue):
        while True:
            task = iq.get()
            if task is None:
                break

            items = task if batch_size > 1 else (task,)
            for arg in items:
                for y in fn(arg):
                    oq.put(y)

        oq.put(None)

    def loader(gen, iq: multiprocessing.Queue):
        batch = []
        for gen_item in gen:
            if batch_size == 1:
                iq.put(gen_item)
            else:
                batch.append(gen_item)
                if len(batch) == batch_size:
                    iq.put(batch)
                    batch = []

        if batch:
            iq.put(batch)

        for _ in range(threads):
            iq.put(None)

    loader_thread = threading.Thread(target=loader, args=(generator, in_q), daemon=True)
    loader_thread.start()

    workers = [multiprocessing.Process(target=worker, args=(in_q, out_q), daemon=True) for _ in range(threads)]
    for p in workers:
        p.start()

    finished = 0
    while finished < threads:
        item = out_q.get()
        if item is None:
            finished += 1
        else:
            yield item

    loader_thread.join()
    for p in workers:
        p.join()


def step(description: str):
    def decorator(method):
        _, line_no = inspect.getsourcelines(method)
        return PipelineActivity.Step(method, line_no, description)

    return decorator


class SkipException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class StopException(KeyboardInterrupt):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def _pp_time(elapsed):
    elapsed = int(elapsed)
    parts = []

    if elapsed > 86400:  # days
        d = int(elapsed / 86400)
        elapsed -= d * 86400
        parts.append(f'{d}d')
    if elapsed > 3600:  # hours
        h = int(elapsed / 3600)
        elapsed -= h * 3600
        parts.append(f'{h}h')
    if elapsed > 60:  # minutes
        m = int(elapsed / 60)
        elapsed -= m * 60
        parts.append(f'{m}m')
    parts.append(f'{elapsed}s')
    return ' '.join(parts)


class PipelineActivity:
    class Step:
        def __init__(self, f: MethodType, line_no: int, description: str):
            self.id: str = f.__name__.strip('_')
            self.name: str = f.__name__
            self.declaring_class: str = f.__qualname__.split('.')[0]
            self._description: str = description
            self._line_no: int = line_no
            self._f: MethodType = f

            sig = inspect.signature(f)
            self.arg_count: int = len([
                param for param in sig.parameters.values()
                if param.kind in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD)
            ]) - 1  # subtract 1 for 'self'

        def __lt__(self, other: 'PipelineActivity.Step') -> bool:
            return self._line_no < other._line_no

        def invoke(self, activity: 'PipelineActivity', input_path: Optional[str], output_path: Optional[str]):
            if self.arg_count == 1:
                self._f(activity, input_path)
            else:
                self._f(activity, input_path, output_path)

        def __str__(self) -> str:
            return self._description

        def __repr__(self) -> str:
            return f'Step(line={self._line_no}, id={self.id}, desc={self._description})'

    @classmethod
    def steps(cls) -> List['PipelineActivity.Step']:
        return sorted([method for _, method in inspect.getmembers(cls)
                       if isinstance(method, PipelineActivity.Step) and method.declaring_class == cls.__name__])

    def __init__(self, args, *, extra_argv=None, input_path: str = None, output_path: str = None, wdir: str = None,
                 log_file: Union[str, TextIOWrapper] = None, start_step: int = None, delete_on_exit: bool = True):
        self.args = args
        self.extra_argv = extra_argv
        self.delete_on_exit: bool = delete_on_exit
        self._input_path: Optional[str] = input_path
        self._output_path: Optional[str] = output_path

        # Configuring working dir
        if wdir is None:
            _wdir = _temp_dir = tempfile.mkdtemp(prefix='larakit_tmp_')
        else:
            _wdir = wdir
            _temp_dir = None

        self._wdir: str = _wdir
        self._temp_dir: Optional[str] = _temp_dir

        if not os.path.isdir(self._wdir):
            os.makedirs(self._wdir)

        # Configuring logging file
        self._close_log_on_exit: bool = False

        self.log_file = log_file or os.path.join(self._wdir, self.__class__.__name__ + '.log')

        if isinstance(self.log_file, str):
            # pylint:disable=consider-using-with
            _log_fobj = open(self.log_file, 'a', encoding='utf-8')
            self._close_log_on_exit = True
        else:
            _log_fobj = self.log_file

        self._log_fobj: TextIOWrapper = _log_fobj

        logging.basicConfig(format='%(asctime)-15s [%(levelname)s] - %(message)s',
                            level=logging.INFO, stream=self._log_fobj)
        self._logger: Logger = logging.getLogger(type(self).__name__)

        # Resuming state
        self._steps: List['PipelineActivity.Step'] = self.steps()
        self.state: StatefulNamespace = StatefulNamespace(os.path.join(self._wdir, 'state.json'), autosave=False,
                                                          step_no=-1)

        if start_step is not None:
            self.state.step_no = start_step

    @property
    def log_fobj(self) -> TextIOWrapper:
        return self._log_fobj

    def wdir(self, *paths: str, create_if_not_exists: bool = True) -> str:
        path = os.path.abspath(os.path.join(self._wdir, *paths))
        if create_if_not_exists and not os.path.isdir(path):
            os.makedirs(path)

        return path

    def native_logging(self, stream=sys.stderr) -> ContextManager[Any]:
        class _NativeLogger:
            def __init__(self, from_fd, to_fd):
                self._from_fd = from_fd
                self._to_fd = to_fd
                self._bak_fd = None

            def __enter__(self):
                self._bak_fd = os.dup(self._from_fd)
                os.dup2(self._to_fd, self._from_fd)
                return self

            def __exit__(self, *args):
                os.dup2(self._bak_fd, self._from_fd)

        return _NativeLogger(stream.fileno(), self._log_fobj.fileno())

    def _index_of_step(self, step_id: str) -> Optional[int]:
        for i, s in enumerate(self._steps):
            if s.id == step_id:
                return i
        return None

    def _remove_step(self, step_id: str) -> None:
        idx = self._index_of_step(step_id)
        if idx is not None:
            del self._steps[idx]

    def save_state(self) -> None:
        self.state.save()

    def run(self) -> None:
        try:
            self.state.pipeline = self.state.pipeline or []

            for i, current_step in enumerate(self._steps):
                step_desc = f'({i + 1}/{len(self._steps)}) {current_step}...'

                print(f'{step_desc:<65s}', end='', flush=True)

                if self.state.step_no < i:
                    self._run_step(i, current_step)
                    self.state.step_no = i
                    self.save_state()
                else:
                    print('SKIPPED', flush=True)

            if self.delete_on_exit:
                shutil.rmtree(self._wdir, ignore_errors=True)
        finally:
            if self._log_fobj is not None and self._close_log_on_exit:
                self._log_fobj.close()
            if self._temp_dir is not None and os.path.isdir(self._temp_dir):
                shutil.rmtree(self._temp_dir, ignore_errors=True)

    def _run_step(self, i, current_step):
        try:
            self._logger.info('Step "%s" started', current_step.id)
            begin = time.time()

            input_path = self._input_path if len(self.state.pipeline) == 0 else self.state.pipeline[-1]
            output_path = None
            if current_step.arg_count > 1:
                if i == len(self._steps) - 1 and self._output_path:
                    if os.path.exists(self._output_path):
                        shutil.rmtree(self._output_path)
                    os.makedirs(self._output_path)

                    output_path = self._output_path
                else:
                    output_path = self.wdir(f'step_{current_step.name}')

            current_step.invoke(self, input_path, output_path)

            if output_path is not None:
                self.state.pipeline.append(output_path)

            elapsed_time = time.time() - begin
            self._logger.info('Step "%s" completed in %s', current_step.id, _pp_time(elapsed_time))

            print(f'DONE in {_pp_time(elapsed_time)}', flush=True)
        except SkipException:
            self._logger.info('Step "%s" skipped', current_step.id)
            print('SKIPPED', flush=True)



import json
import os
from collections import Counter
from typing import Dict, Optional, Set, TextIO
from typing import Generator, Any

from larakit import shell, LanguageDirection
from larakit.corpus._base import TranslationUnit, Properties, MultilingualCorpus, TUReader, TUWriter


class JTMReader(TUReader):
    def __init__(self, path: str):
        self._path = path
        self._file: Optional[TextIO] = None

    def __enter__(self) -> 'JTMReader':
        self._file = open(self._path, 'r', encoding='utf-8')
        return self

    def __iter__(self) -> Generator[TranslationUnit, None, None]:
        if self._file is None:
            raise IOError("Reader is not open.")

        for line in self._file:
            if not line.startswith(JTMCorpus.Footer.FOOTER_LINE_BEGIN):
                yield TranslationUnit.from_json(json.loads(line))

    def __exit__(self, exc_type, exc_value, traceback):
        self._file.close()


class JTMWriter(TUWriter):
    def __init__(self, path: str, properties: Optional[Properties] = None):
        self._path = path
        self._file: Optional[TextIO] = None
        self._properties = properties
        self._counter: Counter[LanguageDirection] = Counter()

    def __enter__(self) -> 'JTMWriter':
        self._file = open(self._path, 'w', encoding='utf-8')
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._file.write(f'{JTMCorpus.Footer(self._counter, self._properties)}\n')
        self._file.close()

    def add_property(self, key: str, value: str):
        if self._properties is None:
            self._properties = Properties()

        self._properties.put(key, value)

    def write(self, tu: TranslationUnit):
        self._counter[tu.language] += 1
        self._file.write(json.dumps(tu.to_json(), ensure_ascii=False, separators=(',', ':')) + '\n')


class JTMCorpus(MultilingualCorpus):
    class Footer:
        FOOTER_LINE_BEGIN = ".footer"

        def __init__(self, counter: Optional[Counter[LanguageDirection]] = None,
                     properties: Optional[Properties] = None):
            self._counter: Counter[LanguageDirection] = counter if counter else Counter()
            self._properties: Optional[Properties] = properties

        @classmethod
        def parse(cls, line: str) -> 'JTMCorpus.Footer':
            if line.startswith(cls.FOOTER_LINE_BEGIN):
                json_part = line[len(cls.FOOTER_LINE_BEGIN):].strip()
                return cls.from_json(json.loads(json_part))

            raise ValueError("The final line does not start with the expected '.footer' prefix.")

        @classmethod
        def from_json(cls, data: Dict[str, Any]) -> 'JTMCorpus.Footer':
            counter = Counter({LanguageDirection.from_tuple(lang_tuple): lang_count
                               for lang_tuple, lang_count in data.get("counter", [])})

            properties_data: Optional[Dict] = data.get("properties", None)
            properties = Properties.from_json(properties_data) if properties_data else None
            return cls(counter, properties)

        @property
        def counter(self) -> Counter[LanguageDirection]:
            return self._counter

        @property
        def properties(self) -> Properties:
            return self._properties

        def get_total_count(self) -> int:
            return self._counter.total()

        def to_json(self) -> Dict[str, Any]:
            data: Dict[str, Any] = {
                "counter": [[lang.to_json(), count] for lang, count in self.counter.items()]}
            if self._properties:
                data["properties"] = self._properties.map

            return data

        def __str__(self) -> str:
            return f"{JTMCorpus.Footer.FOOTER_LINE_BEGIN}{json.dumps(self.to_json(), separators=(',', ':'))}"

    def __init__(self, path: str):
        self._path: str = path
        self._name: str = os.path.basename(path)
        self._footer: Optional[JTMCorpus.Footer] = None

    @property
    def path(self) -> str:
        return self._path

    @property
    def name(self) -> str:
        return self._name

    @property
    def languages(self) -> Set[LanguageDirection]:
        return set(self.footer.counter.keys())

    @property
    def footer(self) -> 'JTMCorpus.Footer':
        if self._footer is None:
            self._footer = self._parse_footer()

        return self._footer

    @property
    def properties(self) -> Optional[Properties]:
        return self.footer.properties if self.footer else None

    def _parse_footer(self) -> Optional['JTMCorpus.Footer']:
        if not os.path.exists(self._path):
            return None

        last_line = shell.tail_1(self._path).decode("utf-8")
        return JTMCorpus.Footer.parse(last_line)

    def reader(self) -> JTMReader:
        return JTMReader(self._path)

    def writer(self, properties: Optional[Properties] = None) -> JTMWriter:
        return JTMWriter(self._path, properties or self.properties)

    def __len__(self):
        return self.footer.get_total_count() if self.footer else 0



import sys
import threading
import time
from typing import Optional


class Color:
    black: 'Color'
    red: 'Color'
    green: 'Color'
    yellow: 'Color'
    blue: 'Color'
    magenta: 'Color'
    cyan: 'Color'
    white: 'Color'

    def __init__(self, code: int) -> None:
        self.__code: int = code

    @property
    def foreground(self) -> str:
        return f'\033[3{self.__code}m'

    @property
    def background(self) -> str:
        return f'\033[4{self.__code}m'


Color.black = Color(0)
Color.red = Color(1)
Color.green = Color(2)
Color.yellow = Color(3)
Color.blue = Color(4)
Color.magenta = Color(5)
Color.cyan = Color(6)
Color.white = Color(7)


class Progressbar:
    def __init__(self, label: str = None,
                 bar_length: int = 40, refresh_time_in_seconds: float = 1., color: Color = Color.white):
        self._is_tty: bool = sys.stdout.isatty()

        # style
        self.label: str = label

        if self._is_tty:
            self._color: str = color.foreground
            self._bar_fill: str = '█'
            self._bar_empty: str = '░'
            self._line_begin: str = '\r'
            self._line_end: str = ''

            self._esc_color_end: str = '\033[m'
            self._esc_hide_cursor: str = '\033[?25l'
            self._esc_show_cursor: str = '\033[?25h'
        else:
            self._color: str = ''
            self._bar_fill: str = '.'
            self._bar_empty: str = ' '
            self._line_begin: str = ''
            self._line_end: str = '\n'

            self._esc_color_end: str = ''
            self._esc_hide_cursor: str = ''
            self._esc_show_cursor: str = ''

        # state
        self._refresh_timeout: float = refresh_time_in_seconds
        self._bar_length: int = bar_length

        self._start_time: Optional[float] = None
        self._rendered_progress: float = -1.0
        self._progress: float = 0.0
        self._background_thread: Optional[threading.Timer] = None
        self._previous_update_length: int = 0

    def _timer_handle(self) -> None:
        self._update()
        self._background_thread = threading.Timer(self._refresh_timeout, self._timer_handle)
        self._background_thread.start()

    def _render(self, elapsed_time: float, eta: float, message: str = None) -> str:
        bar_fill_len = int(self._bar_length * self._progress)
        bar_text = (self._color +
                    (self._bar_fill * bar_fill_len) + (self._bar_empty * (self._bar_length - bar_fill_len)) +
                    self._esc_color_end)

        progress_text = f'{round(100.0 * self._progress, 1):.1f}%'.rjust(6)

        elapsed_time = int(round(elapsed_time))
        eta = int(round(eta))

        elapsed_text = f'{int(elapsed_time / 60):02d}:{elapsed_time % 60:02d}'
        eta_text = f'ETA {int(eta / 3600):02d}:{int(eta / 60) % 60:02d}:{eta % 60:02d}' if eta > 0 else None

        label = None
        if self.label is not None:
            label = '- ' + self.label

            if self._progress < 1:
                label += '...'
            else:
                label += ' completed.'

        message = None if message is None else ' ' + message

        elements = [progress_text, bar_text, elapsed_text, eta_text, label, message]
        return self._line_begin + ' '.join(x for x in elements if x is not None) + self._line_end

    def _should_update(self) -> bool:
        if self._rendered_progress < 0 or self._is_tty:
            return True
        if self._progress == self._rendered_progress:
            return False
        if self._progress in (0, 1):
            return True

        return self._progress // 0.1 != self._rendered_progress // 0.1

    def _update(self, message: str = None) -> None:
        if message is None and not self._should_update():
            return

        elapsed = time.time() - self._start_time
        eta = elapsed / self._progress - elapsed if self._progress > 0 else 0

        update_string = self._render(elapsed, eta, message) + self._esc_hide_cursor

        if self._is_tty:
            if len(update_string) < self._previous_update_length:
                update_string = update_string + ' ' * (self._previous_update_length - len(update_string))
            self._previous_update_length = len(update_string)

        sys.stdout.write(update_string)
        sys.stdout.flush()

        self._rendered_progress = self._progress

    def start(self) -> None:
        self._start_time = time.time()
        self._timer_handle()

    def set_progress(self, progress: float) -> None:
        self._progress = min(1., max(0., progress))

    def cancel(self) -> None:
        if self._background_thread is not None:
            self._background_thread.cancel()
        sys.stdout.write(self._esc_show_cursor)

    def complete(self) -> None:
        if self._background_thread is not None:
            self._background_thread.cancel()
        self._progress = 1.0
        self._update()
        sys.stdout.write('\n')
        sys.stdout.write(self._esc_show_cursor)

    def abort(self, error: str = None) -> None:
        if self._background_thread is not None:
            self._background_thread.cancel()
        self._update(message=None if error is None else f' ERROR: {error}')
        sys.stdout.write('\n')
        sys.stdout.write(self._esc_show_cursor)



import json
import os
from typing import Dict, Any


class Namespace:
    @classmethod
    def from_json_string(cls, json_string: str) -> 'Namespace':
        return cls(**json.loads(json_string))

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            self.set(key, value)

    def get(self, key, default=None):
        return self.__dict__.get(key, default)

    def __getattr__(self, key):
        return self.get(key)

    def _set_private(self, key, value):
        if not key.startswith('_'):
            raise KeyError(f'Private properties must start with an underscore: {key}')
        super().__setattr__(key, value)

    def set(self, key, value):
        if key.startswith('_'):
            raise KeyError(f'Private properties are not allowed: {key}')

        def parse(val):
            if isinstance(val, dict):
                return Namespace(**val)
            if isinstance(val, list):
                return [parse(v) for v in val]
            return val

        super().__setattr__(key, parse(value))

    def __setattr__(self, key, value):
        self.set(key, value)

    def has(self, key):
        return key in self.__dict__

    def __contains__(self, key):
        return self.has(key)

    def __repr__(self):
        return f'{self.__class__.__name__}{self.__dict__}'

    def __str__(self):
        return str(self.to_json())

    def to_json(self) -> Dict[str, Any]:
        def _to_json(value):
            if isinstance(value, Namespace):
                return value.to_json()
            if isinstance(value, list):
                return [_to_json(v) for v in value]
            return value

        return {key: _to_json(value) for key, value in self.__dict__.items() if not key.startswith('_')}


class StatefulNamespace(Namespace):
    def __init__(self, path: str, autosave: bool = False, **default_kwargs):
        super().__init__()
        self._set_private('_path', path)
        self._set_private('_autosave', autosave)

        for key, value in default_kwargs.items():
            super().set(key, value)

        if os.path.isfile(self._path):
            with open(self._path, 'r', encoding='utf-8') as f_input:
                data = json.load(f_input)
                for key, value in data.items():
                    super().set(key, value)

        if self._autosave:
            self.save()

    @property
    def path(self) -> str:
        return self._path

    @property
    def autosave(self) -> bool:
        return self._autosave

    def set(self, key, value):
        super().set(key, value)

        if self._autosave:
            self.save()

    def save(self, indent: int = 2, sort_keys: bool = True):
        with open(self._path, 'w', encoding='utf-8') as f_output:
            f_output.write(json.dumps(self.to_json(), indent=indent, sort_keys=sort_keys))



import atexit
import logging
import os
import re
import shutil
import subprocess
import unicodedata
from collections.abc import Callable
from typing import BinaryIO, Generator, List, Union, IO, Dict, Tuple, Optional

DEVNULL = open(os.devnull, 'wb')  # pylint:disable=consider-using-with
atexit.register(DEVNULL.close)


class ShellError(Exception):
    def __init__(self, command: str, err_no: int, message: str = None):
        self.command: str = command
        self.errno: int = err_no
        self.message: Optional[str] = message

    def __str__(self):
        string = f"Command '{self.command}' failed with exit code {self.errno}"
        if self.message is not None:
            string += ': ' + repr(self.message)
        return string

    def __repr__(self):
        return self.__str__()


def shexec(cmd: Union[str, List[str]], *, stdin: Union[str, IO] = None,
           stdout: IO = subprocess.PIPE, stderr: IO = subprocess.PIPE, background: bool = False,
           env: Dict[str, str] = None, cwd: str = None) -> Union[Tuple[str, str], subprocess.Popen]:
    str_cmd = cmd if isinstance(cmd, str) else ' '.join(cmd)
    logging.getLogger('shell_exec').info(str_cmd)

    message = None
    if background:
        if stdout == subprocess.PIPE:
            stdout = DEVNULL
        if stderr == subprocess.PIPE:
            stderr = DEVNULL
    elif stdin is not None and isinstance(stdin, str):
        message = stdin
        stdin = subprocess.PIPE

    is_shell = isinstance(cmd, str)
    # pylint:disable=consider-using-with
    process = subprocess.Popen(cmd, stdin=stdin, stdout=stdout, stderr=stderr, shell=is_shell, env=env, cwd=cwd)

    stdout_dump = None
    stderr_dump = None
    return_code = 0

    if message is not None or stdout == subprocess.PIPE or stderr == subprocess.PIPE:
        stdout_dump, stderr_dump = process.communicate(message)
        return_code = process.returncode
    elif not background:
        return_code = process.wait()

    if background:
        return process

    if stdout_dump is not None:
        stdout_dump = stdout_dump.decode('utf-8')
    if stderr_dump is not None:
        stderr_dump = stderr_dump.decode('utf-8')

    if return_code != 0:
        raise ShellError(str_cmd, return_code, stderr_dump)
    return stdout_dump, stderr_dump


def safe_open(path: str, mode: str = 'r', encoding: str = 'utf-8') -> Optional[IO]:
    if path is None:
        return None

    if 'w' in mode or 'a' in mode:
        folder = os.path.dirname(path)
        if not os.path.isdir(folder):
            os.makedirs(folder)
    elif not os.path.isfile(path):
        return None

    return open(path, mode=mode, encoding=encoding)


def sanitize_filename(filename: str, allow_unicode: bool = False) -> str:
    filename = str(filename).strip()

    if allow_unicode:
        filename = unicodedata.normalize('NFKC', filename)
    else:
        filename = (unicodedata.normalize('NFKD', filename)
                    .encode('ascii', 'ignore')
                    .decode('ascii'))

    return re.sub(r'[^\w\s\-.]', '_', filename)


def tail_1(path: str) -> bytes:
    file_size = os.path.getsize(path)

    window_size = 1024

    with open(path, 'rb') as file:
        while True:
            file.seek(-min(file_size, window_size), 2)
            last_line = file.read().rstrip(b'\n')

            if b'\n' in last_line:
                return last_line[last_line.rindex(b'\n') + 1:]
            if window_size >= file_size:
                return last_line
            window_size *= 2


def lc(filename: str, block_size: int = 65536) -> int:
    if not os.path.isfile(filename):
        raise FileNotFoundError(f'File "{filename}" does not exist.')

    def _blocks(files: BinaryIO) -> Generator[bytes, None, None]:
        while True:
            b = files.read(block_size)
            if not b:
                break
            yield b

    with open(filename, 'rb') as stream:
        count = 0
        for block in _blocks(stream):
            count += block.count(b'\n')
        return count


def cat(files: List[str], output: str, *, buffer_size: int = 10 * 1024 * 1024, append: bool = False):
    mode = 'wb' if not append else 'ab'
    with open(output, mode) as blob:
        for f in files:
            with open(f, 'rb') as source:
                shutil.copyfileobj(source, blob, buffer_size)


def link(file_path: str, dest_path: str, symbolic: bool = False, overwrite: bool = True) -> str:
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"File '{file_path}' does not exist.")

    if os.path.isdir(dest_path):
        dest_path = os.path.join(dest_path, os.path.basename(file_path))
    elif os.path.isfile(dest_path):
        if not overwrite:
            raise IOError(f"Destination path '{dest_path}' already exists.")
        os.remove(dest_path)
    else:
        dirname = os.path.dirname(dest_path)
        if not os.path.isdir(dirname):
            raise IOError(f"Destination directory '{dirname}' does not exist.")

    if symbolic:
        os.symlink(file_path, dest_path)
    else:
        os.link(file_path, dest_path)

    return dest_path


def tar_gz(output_archive: str, input_dir: str, *,
           file_filter: Callable[[str], bool] = None, use_pigz: bool = True) -> None:
    filenames = [f for f in os.listdir(input_dir) if file_filter is None or file_filter(os.path.join(input_dir, f))]

    cmd = ['tar', '-cf', output_archive]
    if use_pigz:
        cmd.append('--use-compress-program=pigz')
    cmd.extend(['--directory', input_dir])
    cmd.extend(filenames)

    shexec(cmd)


def untar_gz(input_archive: str, output_dir: str, *, use_pigz: bool = True) -> None:
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    cmd = ['tar', '-xf', input_archive]
    if use_pigz:
        cmd.append('--use-compress-program=pigz')
    cmd.extend(['--directory', output_dir])

    shexec(cmd)


def nvidia_devices() -> List[int]:
    try:
        output, _ = shexec('nvidia-smi --query-gpu=index --format=csv,noheader,nounits')
        return [int(index) for index in output.splitlines()]
    except ShellError:
        return []


def nvidia_device_count() -> int:
    return len(nvidia_devices())


def nvidia_has_device() -> bool:
    return nvidia_device_count() > 0



from typing import Optional, List, Tuple, Dict


class Language:
    _CODE_TO_NAME: Dict[str, str] = {
        "de": "German",
        "en": "English",
        "es": "Spanish",
        "fr": "French",
        "it": "Italian",
        "ja": "Japanese",
        "ko": "Korean",
        "pt": "Portuguese",
        "ru": "Russian",
        "zh": "Chinese",
        "ca": "Catalan",
        "cs": "Czech",
        "da": "Danish",
        "el": "Greek",
        "fi": "Finnish",
        "he": "Hebrew",
        "hr": "Croatian",
        "hu": "Hungarian",
        "id": "Indonesian",
        "nb": "Norwegian Bokmål",
        "nl": "Dutch",
        "vls": "Dutch",
        "nn": "Norwegian Nynorsk",
        "pl": "Polish",
        "sk": "Slovak",
        "sv": "Swedish",
        "tr": "Turkish",
        "ar": "Arabic",
        "bg": "Bulgarian",
        "uk": "Ukrainian",
        "ms": "Malay",
        "th": "Thai",
        "az": "Azerbaijani",
        "bs": "Bosnian",
        "et": "Estonian",
        "ga": "Irish",
        "hi": "Hindi",
        "hy": "Armenian",
        "is": "Icelandic",
        "ka": "Georgian",
        "kn": "Kannada",
        "lt": "Lithuanian",
        "lv": "Latvian",
        "mk": "Macedonian",
        "mr": "Marathi",
        "mt": "Maltese",
        "ro": "Romanian",
        "sl": "Slovenian",
        "sq": "Albanian",
        "sw": "Swahili",
        "sr": "Serbian",
        "tl": "Tagalog",
        "vi": "Vietnamese",
        "xh": "Xhosa",
        "zu": "Zulu",
        "ace": "Acehnese",
        "af": "Afrikaans",
        "ak": "Akan",
        "als": "Tosk Albanian",
        "am": "Amharic",
        "as": "Assamese",
        "ast": "Asturian",
        "awa": "Awadhi",
        "ayr": "Aymara",
        "azb": "Southern Azerbaijani",
        "ba": "Bashkir",
        "ban": "Balinese",
        "be": "Belarusian",
        "bem": "Bemba",
        "bho": "Bhojpuri",
        "bjn": "Banjar",
        "bm": "Bambara",
        "bn": "Bengali",
        "bo": "Tibetan",
        "bug": "Buginese",
        "ceb": "Cebuano",
        "cjk": "Chokwe",
        "ckb": "Central Kurdish",
        "crh": "Crimean Tatar",
        "cy": "Welsh",
        "dik": "Dinka",
        "diq": "Dimli",
        "dyu": "Dyula",
        "dz": "Dzongkha",
        "ee": "Ewe",
        "eo": "Esperanto",
        "eu": "Basque",
        "fa": "Persian",
        "fil": "Filipino",
        "fj": "Fijian",
        "fo": "Faroese",
        "fon": "Fon",
        "fur": "Friulian",
        "fuv": "Fulfulde",
        "gaz": "Oromo",
        "gd": "Gaelic",
        "gl": "Galician",
        "gn": "Guarani",
        "gu": "Gujarati",
        "ha": "Hausa",
        "hne": "Chhattisgarhi",
        "ht": "Haitian",
        "ig": "Igbo",
        "ilo": "Iloko",
        "jv": "Javanese",
        "kab": "Kabyle",
        "kac": "Jingpho",
        "kam": "Kamba",
        "kas": "Arabic Kashmiri",
        "kbp": "Kabiyè",
        "kea": "Kabuverdianu",
        "kg": "Kongo",
        "khk": "Halh Mongolian",
        "ki": "Kikuyu",
        "kk": "Kazakh",
        "km": "Khmer",
        "kmb": "Kimbundu",
        "kmr": "Northern Kurdish",
        "knc": "Kanuri",
        "ks": "Devanagari Kashmiri",
        "ky": "Kyrgyz",
        "la": "Latin",
        "lb": "Luxembourgish",
        "lg": "Ganda",
        "li": "Limburgish",
        "lij": "Ligurian",
        "lmo": "Lombard",
        "ln": "Lingala",
        "lo": "Lao",
        "ltg": "Latgalian",
        "lua": "Luba-Kasai",
        "luo": "Luo",
        "lus": "Mizo",
        "mag": "Magahi",
        "mai": "Maithili",
        "mg": "Malagasy",
        "mi": "Maori",
        "min": "Minangkabau",
        "ml": "Malayalam",
        "mn": "Traditional Mongolian",
        "mni": "Manipuri",
        "mos": "Mossi",
        "my": "Burmese",
        "ne": "Nepali",
        "nso": "Northern Sotho",
        "nus": "Nuer",
        "ny": "Nyanja",
        "oc": "Occitan",
        "or": "Oriya",
        "pa": "Punjabi",
        "pag": "Pangasinan",
        "pap": "Papiamento",
        "pbt": "Southern Pashto",
        "plt": "Plateau Malagasy",
        "prs": "Dari",
        "ps": "Pashto",
        "quy": "Ayacucho Quechua",
        "rn": "Rundi",
        "rw": "Kinyarwanda",
        "sa": "Sanskrit",
        "sat": "Santali",
        "sc": "Sardinian",
        "scn": "Sicilian",
        "sd": "Sindhi",
        "sg": "Sango",
        "shn": "Shan",
        "si": "Sinhala",
        "sm": "Samoan",
        "sn": "Shona",
        "so": "Somali",
        "ss": "Swati",
        "st": "Southern Sotho",
        "su": "Sundanese",
        "szl": "Silesian",
        "ta": "Tamil",
        "taq": "Tamasheq",
        "te": "Telugu",
        "tg": "Tajik",
        "ti": "Tigrinya",
        "tk": "Turkmen",
        "tn": "Tswana",
        "tpi": "Tok Pisin",
        "ts": "Tsonga",
        "tt": "Tatar",
        "tum": "Tumbuka",
        "tw": "Twi",
        "tzm": "Tamazight",
        "ug": "Uyghur",
        "umb": "Umbundu",
        "ur": "Urdu",
        "uzn": "Uzbek",
        "vec": "Venetian",
        "war": "Waray",
        "wo": "Wolof",
        "ydd": "Yiddish",
        "yo": "Yoruba",
    }

    def __init__(self, code: str, tag: str, region: Optional[str] = None, script: Optional[str] = None):
        if not code or not isinstance(code, str):
            raise ValueError("code must be a non-empty string")
        if not tag or not isinstance(tag, str):
            raise ValueError("tag must be a non-empty string")

        self._code: str = code
        self._tag: str = tag
        self._region: Optional[str] = region
        self._script: Optional[str] = script

    @staticmethod
    def _to_title_case(string: str) -> str:
        return string[0].upper() + string[1:].lower()

    @staticmethod
    def _is_alpha(string: str) -> bool:
        return all('a' <= c <= 'z' or 'A' <= c <= 'Z' for c in string)

    @staticmethod
    def _is_digit(string: str) -> bool:
        return all('0' <= c <= '9' for c in string)

    @staticmethod
    def _parse_language(string: str) -> Optional[str]:
        if (len(string) == 2 or len(string) == 3) and Language._is_alpha(string):
            return string.lower()
        return None

    @staticmethod
    def _parse_script(string: str) -> Optional[str]:
        if len(string) == 4 and Language._is_alpha(string):
            return Language._to_title_case(string)
        return None

    @staticmethod
    def _parse_region(string: str) -> Optional[str]:
        if len(string) == 2 and Language._is_alpha(string):
            return string.upper()
        if len(string) == 3 and Language._is_digit(string):
            return string
        return None

    @classmethod
    def from_string(cls, string: str) -> 'Language':
        if string is None:
            raise ValueError("Input string must not be None")
        if not string:
            raise ValueError("Input string must not be empty")

        code: Optional[str] = None
        region: Optional[str] = None
        script: Optional[str] = None
        tag_parts: List[str] = []

        skip: bool = False
        chunks: List[str] = string.replace('_', '-').split('-')
        for i, chunk in enumerate(chunks):
            if i == 0:
                code = Language._parse_language(chunk)
                if code is None:
                    raise ValueError(f"Invalid language subtag: '{chunk}' in '{string}'")
                tag_parts.append(code)
            else:
                if not skip:
                    if i == 1 and script is None:
                        script_candidate = Language._parse_script(chunk)
                        if script_candidate is not None:
                            script = script_candidate
                            tag_parts.append(script)
                            continue

                    if i <= 2 and region is None:
                        region_candidate = Language._parse_region(chunk)
                        if region_candidate is not None:
                            region = region_candidate
                            tag_parts.append(region)
                            continue

                skip = True
                tag_parts.append(chunk)

        if code is None:
            raise ValueError(f"Language subtag missing in '{string}'")

        tag: str = '-'.join(tag_parts)
        return cls(code=code, tag=tag, region=region, script=script)

    @property
    def code(self) -> str:
        return self._code

    @property
    def tag(self) -> str:
        return self._tag

    @property
    def region(self) -> Optional[str]:
        return self._region

    @property
    def script(self) -> Optional[str]:
        return self._script

    @property
    def name(self) -> Optional[str]:
        return self._CODE_TO_NAME.get(self.code, None)

    def is_language_only(self) -> bool:
        return self.code == self.tag

    def as_language_only(self) -> 'Language':
        return self if self.is_language_only() else Language.from_string(self.code)

    def is_equal_or_more_generic_than(self, other: 'Language') -> bool:
        if not other:
            return False

        if self.code != other.code:
            return False
        if self.script and self.script != other.script:
            return False
        if self.region and self.region != other.region:
            return False

        return True

    def __eq__(self, other: 'Language') -> bool:
        return self.tag == other.tag

    def __hash__(self) -> int:
        return hash(self.tag)

    def __repr__(self) -> str:
        return self.tag

    def __str__(self):
        return self.tag


class LanguageDirection:
    def __init__(self, source: Language, target: Language):
        self._source: Language = source
        self._target: Language = target
        self._reversed: Optional['LanguageDirection'] = None

    @property
    def source(self) -> Language:
        return self._source

    @property
    def target(self) -> Language:
        return self._target

    @property
    def reversed(self):
        if self._reversed is None:
            self._reversed = LanguageDirection(source=self.target, target=self.source)
        return self._reversed

    @classmethod
    def from_tuple(cls, language: Tuple[str, str]) -> 'LanguageDirection':
        if len(language) != 2:
            raise ValueError(f"Language tuple must contain two elements, got {len(language)}")
        return cls(source=Language.from_string(language[0]), target=Language.from_string(language[1]))

    def is_equal_or_more_generic_than(self, other: 'LanguageDirection') -> bool:
        return (self.source.is_equal_or_more_generic_than(other.source) and
                self.target.is_equal_or_more_generic_than(other.target))

    def to_json(self) -> Tuple[str, str]:
        return self.source.tag, self.target.tag

    def __eq__(self, other: object) -> bool:
        if self is other:
            return True
        if not isinstance(other, LanguageDirection):
            return False
        return self.source == other.source and self.target == other.target

    def __hash__(self) -> int:
        return hash((self.source, self.target))

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"{self.source}\u2192{self.target}"



from abc import abstractmethod, ABC
from typing import Dict, List, Union, Optional, Iterable, Set, Tuple, Generator

from larakit import LanguageDirection


class Properties:
    def __init__(self, other: Optional['Properties'] = None):
        self._map: Dict[str, Union[str, List[str]]] = {}
        if other:
            self._set_all(other.map)

    @classmethod
    def from_json(cls, json_data: Dict[str, Union[str, List[str]]]) -> 'Properties':
        properties = cls()
        properties._set_all(json_data)
        return properties

    @property
    def map(self) -> Dict[str, Union[str, List[str]]]:
        return self._map

    def put(self, key: str, value: str) -> None:
        existing = self.map.get(key)
        if existing is None:
            self.map[key] = value
        elif isinstance(existing, list):
            existing.append(value)
        else:
            self.map[key] = [existing, value]

    def _set(self, key: str, value: Union[str, List[str]]) -> None:
        self.map[key] = value.copy() if isinstance(value, list) else value

    def _set_all(self, data: Dict[str, Union[str, List[str]]]) -> None:
        for key, value in data.items():
            self._set(key, value)

    def set(self, key: str, value: str) -> None:
        self._set(key, value)

    def has(self, key: str) -> bool:
        return key in self.map

    def keys(self) -> Iterable[str]:
        return self.map.keys()

    def get(self, key: str) -> Optional[Union[str, List[str]]]:
        return self.map.get(key)

    def size(self) -> int:
        return len(self.map)

    def value(self, key: str) -> Optional[str]:
        value = self.map.get(key)
        if value is None:
            return None
        if isinstance(value, list):
            return value[0]
        return value

    def values(self, key: str) -> Optional[List[str]]:
        value = self.map.get(key)
        if value is None:
            return None
        if isinstance(value, list):
            return value[:]
        return [value]

    def remove(self, key: str) -> None:
        self.map.pop(key, None)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Properties):
            return False
        return self.map == other.map

    def __hash__(self) -> int:
        items = tuple(sorted((k, tuple(v) if isinstance(v, list) else v) for k, v in self.map.items()))
        return hash(items)

    def __str__(self) -> str:
        return str(self.map)


class TranslationUnit:
    def __init__(self, language: LanguageDirection, sentence: str, translation: str, *,
                 tuid: Optional[str] = None, creation_date: Optional[str] = None, change_date: Optional[str] = None,
                 properties: Optional[Properties] = None):
        self._language: LanguageDirection = language
        self._sentence: str = sentence
        self._translation: str = translation
        self._tuid: Optional[str] = tuid
        self._creation_date: Optional[str] = creation_date
        self._change_date: Optional[str] = change_date
        self._properties: Optional[Properties] = properties

    @classmethod
    def from_tu(cls, tu: 'TranslationUnit') -> 'TranslationUnit':
        return cls(language=tu.language, sentence=tu.sentence, translation=tu.translation, tuid=tu.tuid,
                   creation_date=tu.creation_date, change_date=tu.change_date,
                   properties=None if tu.properties is None else Properties(tu.properties))

    @classmethod
    def from_json(cls, json_data: Dict[str, Union[str, List[str], Tuple[str, str]]]) -> 'TranslationUnit':
        properties_data: Optional[Dict] = json_data.get("properties", None)
        return cls(language=LanguageDirection.from_tuple(json_data['language']), sentence=json_data["sentence"],
                   translation=json_data["translation"], tuid=json_data.get("tuid", None),
                   creation_date=json_data.get("creationDate", None), change_date=json_data.get("changeDate", None),
                   properties=Properties.from_json(properties_data) if properties_data is not None else None)

    @property
    def language(self) -> LanguageDirection:
        return self._language

    @property
    def sentence(self) -> str:
        return self._sentence

    @property
    def translation(self) -> str:
        return self._translation

    @property
    def tuid(self) -> Optional[str]:
        return self._tuid

    @property
    def creation_date(self) -> Optional[str]:
        return self._creation_date

    @property
    def change_date(self) -> Optional[str]:
        return self._change_date

    @property
    def properties(self) -> Optional[Properties]:
        return self._properties

    def to_json(self) -> Dict[str, Union[str, Tuple[str, str], Dict[str, Union[str, List[str]]]]]:
        data: Dict[str, Union[str, Tuple[str, str], Dict[str, Union[str, List[str]]]]] = {
            "language": self.language.to_json(), "sentence": self.sentence, "translation": self.translation
        }

        if self.tuid is not None:
            data["tuid"] = self.tuid
        if self.creation_date is not None:
            data["creationDate"] = self.creation_date
        if self.change_date is not None:
            data["changeDate"] = self.change_date
        if self.properties is not None:
            data["properties"] = self.properties.map

        return data

    def __eq__(self, other: 'TranslationUnit') -> bool:
        if not isinstance(other, TranslationUnit):
            return False
        return (self.tuid == other.tuid and self.language == other.language and
                self.sentence == other.sentence and self.translation == other.translation and
                self.creation_date == other.creation_date and self.change_date == other.change_date)

    def __hash__(self) -> int:
        return hash((self.tuid, self.language, self.sentence, self.translation, self.creation_date, self.change_date))

    def __str__(self) -> str:
        return f'[{self.language}] <{self.sentence}> ||| <{self.translation}>'


class TUReader(ABC):
    @abstractmethod
    def __enter__(self) -> 'TUReader':
        pass

    @abstractmethod
    def __iter__(self) -> Generator[TranslationUnit, None, None]:
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        pass


class TUWriter(ABC):
    @abstractmethod
    def __enter__(self) -> 'TUWriter':
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        pass

    @abstractmethod
    def write(self, tu: TranslationUnit):
        pass


class MultilingualCorpus(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def languages(self) -> Set[LanguageDirection]:
        pass

    @abstractmethod
    def reader(self) -> TUReader:
        pass

    @abstractmethod
    def writer(self) -> TUWriter:
        pass



import os
from typing import Set, Optional, TextIO, Generator

from larakit import LanguageDirection
from larakit.corpus._base import MultilingualCorpus, TUReader, TranslationUnit, TUWriter


class ParallelCorpusReader(TUReader):
    def __init__(self, language: LanguageDirection, source: str, target: str):
        self._language: LanguageDirection = language
        self._source: str = source
        self._target: str = target

        self._source_file: Optional[TextIO] = None
        self._target_file: Optional[TextIO] = None

    def __enter__(self) -> 'ParallelCorpusReader':
        self._source_file = open(self._source, 'r', encoding='utf-8')
        self._target_file = open(self._target, 'r', encoding='utf-8')
        return self

    def __iter__(self) -> Generator[TranslationUnit, None, None]:
        if self._source_file is None or self._target_file is None:
            raise IOError("Reader is not open.")

        for src_line, tgt_line in zip(self._source_file, self._target_file):
            yield TranslationUnit(language=self._language, sentence=src_line.strip(), translation=tgt_line.strip())

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._source_file.close()
        self._target_file.close()


class ParallelCorpusWriter(TUWriter):
    def __init__(self, language: LanguageDirection, source: str, target: str):
        self._language: LanguageDirection = language
        self._source: str = source
        self._target: str = target

        self._source_file: Optional[TextIO] = None
        self._target_file: Optional[TextIO] = None

    def __enter__(self) -> 'ParallelCorpusWriter':
        self._source_file = open(self._source, 'w', encoding='utf-8')
        self._target_file = open(self._target, 'w', encoding='utf-8')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._source_file:
            self._source_file.close()
        if self._target_file:
            self._target_file.close()

    @staticmethod
    def _normalize(text: str) -> str:
        return text.replace('\n', ' ').strip()

    def write(self, tu: TranslationUnit):
        self._source_file.write(self._normalize(tu.sentence) + '\n')
        self._target_file.write(self._normalize(tu.translation) + '\n')


class ParallelCorpus(MultilingualCorpus):
    def __init__(self, source: str, target: str):
        self._source: str = source
        self._target: str = target

        source_parts = os.path.splitext(os.path.basename(source))
        target_parts = os.path.splitext(os.path.basename(target))
        self._name = source_parts[0]
        self._language: LanguageDirection = LanguageDirection.from_tuple((source_parts[1][1:], target_parts[1][1:]))

    @property
    def name(self) -> str:
        return self._name

    @property
    def languages(self) -> Set[LanguageDirection]:
        return {self._language}

    def reader(self) -> ParallelCorpusReader:
        return ParallelCorpusReader(self._language, self._source, self._target)

    def writer(self) -> ParallelCorpusWriter:
        return ParallelCorpusWriter(self._language, self._source, self._target)



import math
import random
from dataclasses import dataclass, field
from typing import Iterable, Any, List


@dataclass
class Sequence:
    sum: float = field(init=False, default=0.0)
    sum2: float = field(init=False, default=0.0)
    length: int = field(init=False, default=0)

    def add(self, value: float) -> None:
        self.sum += value
        self.sum2 += value ** 2
        self.length += 1

    def merge(self, other: "Sequence") -> None:
        self.sum += other.sum
        self.sum2 += other.sum2
        self.length += other.length

    @property
    def mean(self) -> float:
        if self.length == 0:
            return math.nan
        return self.sum / self.length

    @property
    def stddev(self) -> float:
        if self.length <= 1:
            return math.nan

        avg = self.sum / self.length
        stddev = math.sqrt((self.sum2 / self.length) - (avg ** 2))
        return stddev


def reservoir_sampling(stream: Iterable[Any], size: int) -> List[Any]:
    result = []
    for i, tu in enumerate(stream):
        if i < size:
            result.append(tu)
        else:
            j = random.randint(0, i)
            if j < size:
                result[j] = tu

    return result



import io
import os
from dataclasses import dataclass
from typing import Generator, Optional, Set, TextIO, List, Dict, Tuple, Iterator
from xml.etree import ElementTree as ET
from xml.sax.saxutils import XMLGenerator
from xml.sax.xmlreader import AttributesImpl

from larakit import LanguageDirection, Language
from larakit.corpus._base import MultilingualCorpus, TUReader, TUWriter, TranslationUnit, Properties


class TMXReader(TUReader):
    @dataclass
    class _TUVData:
        lang: Optional[str]
        text: Optional[str]
        creation_date: Optional[str]
        change_date: Optional[str]

    def __init__(self, path: str):
        self._path: str = path
        self._file: Optional[TextIO] = None
        self._header_properties: Optional[Properties] = None
        self._header_srclang: Optional[str] = None

    def __enter__(self) -> 'TMXReader':
        self._file = open(self._path, 'r', encoding='utf-8')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._file:
            self._file.close()

    @staticmethod
    def _local_name(tag: str) -> str:
        if tag is None:
            return ""
        if tag.startswith("{"):
            return tag.split("}", 1)[1]
        return tag

    @staticmethod
    def _get_lang(attrib: Dict[str, str]) -> Optional[str]:
        return attrib.get("{http://www.w3.org/XML/1998/namespace}lang") or attrib.get("lang")

    def _parse_header(self, context: Iterator[Tuple[str, ET.Element]]) -> None:
        if self._header_properties:
            return

        self._header_properties = Properties()
        for event, elem in context:
            tag = self._local_name(elem.tag)
            if event == 'start' and tag == 'header':
                self._header_srclang = elem.attrib.get("srclang")
            elif event == 'end' and tag == 'prop' and 'type' in elem.attrib:
                self._header_properties.put(elem.attrib['type'], "".join(elem.itertext()))
            elif event == 'end' and tag == 'header':
                elem.clear()
                return

    @staticmethod
    def _find_source_tuv_index(tuvs: List[_TUVData], source_lang: Optional[Language]) -> int:
        if not source_lang:
            return 0

        tuv_langs: List[Language] = [Language.from_string(tuv.lang) if tuv.lang else None for tuv in tuvs]
        # 1. Exact match
        for i, tuv_lang in enumerate(tuv_langs):
            if tuv_lang == source_lang:
                return i
        # 2. Generic match (e.g., source 'en' matches TUV 'en-US')
        for i, tuv_lang in enumerate(tuv_langs):
            if tuv_lang and source_lang.is_equal_or_more_generic_than(tuv_lang):
                return i
        # 3. Specific match (e.g., source 'en-US' matches TUV 'en')
        for i, tuv_lang in enumerate(tuv_langs):
            if tuv_lang and tuv_lang.is_equal_or_more_generic_than(source_lang):
                return i
        # 4. Fallback to the first TUV
        return 0

    @classmethod
    def _tuvs_from_element(cls, tu_element: ET.Element) -> List[_TUVData]:
        tuvs: List[TMXReader._TUVData] = []
        for tuv_elem in tu_element.findall('tuv'):
            seg_elem = tuv_elem.find('seg')
            if seg_elem is None:
                continue

            tuv = TMXReader._TUVData(lang=cls._get_lang(tuv_elem.attrib), text="".join(seg_elem.itertext()),
                                     creation_date=tuv_elem.attrib.get("creationdate"),
                                     change_date=tuv_elem.attrib.get("changedate"))
            tuvs.append(tuv)
        return tuvs

    def _translation_units_from_element(self, tu_element: ET.Element) -> Generator[TranslationUnit, None, None]:
        tuvs = self._tuvs_from_element(tu_element)
        if len(tuvs) < 2:
            return

        # Determine source language, falling back from TU, header, or first TUV
        source_lang_tag = tu_element.attrib.get("srclang") or self._header_srclang
        if not source_lang_tag and tuvs[0].lang:
            source_lang_tag = tuvs[0].lang

        source_lang = Language.from_string(source_lang_tag) if source_lang_tag else None
        source_tuv = tuvs.pop(self._find_source_tuv_index(tuvs, source_lang))

        # Extract TU-level metadata
        tu_props = Properties()
        for prop_elem in tu_element.findall('prop'):
            prop_type = prop_elem.attrib.get('type')
            if prop_type:
                tu_props.put(prop_type, "".join(prop_elem.itertext()))

        # Yield a TranslationUnit for each remaining target TUV
        for target_tuv in tuvs:
            if not all([source_tuv.lang, source_tuv.text, target_tuv.lang, target_tuv.text]):
                continue

            lang_dir = LanguageDirection.from_tuple((source_tuv.lang, target_tuv.lang))
            yield TranslationUnit(
                language=lang_dir, sentence=source_tuv.text, translation=target_tuv.text,
                tuid=tu_element.attrib.get("tuid"),
                creation_date=target_tuv.creation_date or tu_element.attrib.get("creationdate"),
                change_date=target_tuv.change_date or tu_element.attrib.get("changedate"),
                properties=Properties(tu_props) if tu_props.size() > 0 else None)

    def __iter__(self) -> Generator[TranslationUnit, None, None]:
        if self._file is None:
            raise IOError("Reader is not open.")

        context = ET.iterparse(self._file, events=("start", "end"))
        self._parse_header(context)
        for event, elem in context:
            if event == 'end' and self._local_name(elem.tag) == 'tu':
                yield from self._translation_units_from_element(elem)
                elem.clear()

    @property
    def header_properties(self) -> Optional[Properties]:
        return self._header_properties

    @property
    def header_source_language(self) -> Optional[str]:
        return self._header_srclang


class TMXWriter(TUWriter):
    def __init__(self, path: str, header_properties: Optional[Properties] = None):
        self._path: str = path
        self._file: Optional[TextIO] = None
        self._xml_gen: Optional[XMLGenerator] = None
        self._header_written: bool = False
        self._header_properties = header_properties

    def __enter__(self) -> 'TMXWriter':
        self._file = open(self._path, 'wb')
        text_stream = io.TextIOWrapper(self._file, encoding='utf-8', write_through=True)
        self._xml_gen = XMLGenerator(out=text_stream, encoding='utf-8')
        self._xml_gen.startDocument()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._xml_gen:
            self._xml_gen.endElement("body")
            self._xml_gen.endElement("tmx")
            self._xml_gen.endDocument()
        if self._file:
            self._file.close()

    @staticmethod
    def _sanitize_text(chunk: str) -> str:
        def valid_xml_codepoint(c: int) -> bool:
            return ((c == 0x09) or (c == 0x0A) or (c == 0x0D) or (0x20 <= c <= 0xD7FF) or (0xE000 <= c <= 0xFFFD) or (
                    0x10000 <= c <= 0x10FFFF))

        def clean(text: str) -> str:
            if all(valid_xml_codepoint(ord(ch)) for ch in text):
                return text

            return ''.join(ch for ch in text if valid_xml_codepoint(ord(ch)))

        return clean(chunk)

    def write(self, tu: TranslationUnit) -> None:
        if self._xml_gen is None:
            raise IOError("Writer is not open.")
        if not self._header_written:
            self._write_header(tu.language.source)
            self._header_written = True
        self._write_tu(tu)

    def _write_header(self, srclang: Optional[Language]) -> None:
        self._xml_gen.startElement("tmx", AttributesImpl({"version": "1.4"}))

        header_attrs = {"datatype": "plaintext", "o-tmf": "LaraKit", "segtype": "sentence", "adminlang": "en"}
        if srclang:
            header_attrs["srclang"] = srclang.tag

        self._xml_gen.startElement("header", AttributesImpl(header_attrs))
        if self._header_properties:
            for key in self._header_properties.keys():
                for val in self._header_properties.values(key) or []:
                    self._xml_gen.startElement("prop", AttributesImpl({"type": key}))
                    self._xml_gen.characters(val)
                    self._xml_gen.endElement("prop")
        self._xml_gen.endElement("header")
        self._xml_gen.startElement("body", AttributesImpl({}))

    def _write_tu(self, tu: TranslationUnit) -> None:
        attrs = {"datatype": "plaintext", "srclang": tu.language.source.tag}
        if tu.tuid:
            attrs["tuid"] = tu.tuid
        if tu.creation_date:
            attrs["creationdate"] = tu.creation_date
        if tu.change_date:
            attrs["changedate"] = tu.change_date

        self._xml_gen.startElement("tu", AttributesImpl(attrs))
        if tu.properties is not None:
            for key in tu.properties.keys():
                for val in tu.properties.values(key) or []:
                    self._xml_gen.startElement("prop", AttributesImpl({"type": key}))
                    self._xml_gen.characters(val)
                    self._xml_gen.endElement("prop")

        self._write_tuv(tu.language.source, tu.sentence)
        self._write_tuv(tu.language.target, tu.translation)
        self._xml_gen.endElement("tu")

    def _write_tuv(self, lang: Language, segment: str) -> None:
        self._xml_gen.startElement("tuv", AttributesImpl({"xml:lang": lang.tag}))
        self._xml_gen.startElement("seg", AttributesImpl({}))
        self._xml_gen.characters(self._sanitize_text(segment))
        self._xml_gen.endElement("seg")
        self._xml_gen.endElement("tuv")


class TMXCorpus(MultilingualCorpus):
    def __init__(self, path: str):
        self._path: str = path
        self._name: str = os.path.splitext(os.path.basename(path))[0]
        self._languages: Optional[Set[LanguageDirection]] = None
        self._length: Optional[int] = None
        self._header_properties: Optional[Properties] = None

    @property
    def path(self) -> str:
        return self._path

    @property
    def name(self) -> str:
        return self._name

    @property
    def languages(self) -> Optional[Set[LanguageDirection]]:
        if not os.path.exists(self._path):
            return None

        if self._languages is None:
            langs: Set[LanguageDirection] = set()
            with TMXReader(self._path) as r:
                for tu in r:
                    langs.add(tu.language)
            self._languages = langs

        return self._languages

    def reader(self) -> TMXReader:
        return TMXReader(self._path)

    def writer(self, properties: Optional[Properties] = None) -> TMXWriter:
        return TMXWriter(self._path, properties or self.properties)

    @property
    def properties(self) -> Optional[Properties]:
        if not os.path.exists(self._path):
            return None

        if self._header_properties is not None:
            return self._header_properties
        with TMXReader(self._path) as r:
            for _ in r:
                break
            self._header_properties = r.header_properties if r.header_properties.size() > 0 else None

        return self._header_properties

    def __len__(self) -> int:
        if not os.path.exists(self._path):
            return 0

        if self._length is None:
            with TMXReader(self._path) as r:
                self._length = sum(1 for _ in r)
        return self._length
